/*
 * (C) Copyright IBM Corp. 1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.dom.util;

import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Comment;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.DocumentType;
import org.w3c.dom.Element;
import org.w3c.dom.Entity;
import org.w3c.dom.EntityReference;
import org.w3c.dom.Node;
import org.w3c.dom.Notation;
import org.w3c.dom.ProcessingInstruction;
import org.w3c.dom.Text;
/**
 *
 * @version Revision: %M% %I% %W% %Q%
 * @author Tom Rowe &lt;trowe@us.ibm.com&gt;
 * @author TAMURA Kent &lt;tkent@jp.ibm.com&gt;
 * @see com.ibm.xml.parser.NonRecursivePreorderTreeTraversal
 */
public class TreeTraversal {

    Visitor visitor;

    /**
     *
     */
    public TreeTraversal(Visitor visitor) {
        this.visitor = visitor;
    }
    /**
     * Traverses the document object tree at the specified <var>startNode</var>.
     * <p>The numbered nodes in the trees below indicate the order of traversal given
     * the specified <code>startNode</code> of &quot;1&quot;.
     * <pre>
     *
     *                 1              x              x
     *                / \            / \            / \
     *               2   6          1   x          x   x
     *              /|\   \        /|\   \        /|\   \
     *             3 4 5   7      2 3 4   x      x 1 x   x
     *
     * </pre>
     * @param   startNode   The starting point to begin traversing the document object tree.
     * @exception Exception Thrown if the document hierarchy can not be visitted.
     */
    public void traverse(Node startNode) throws Exception {
        Node tempNode1 = startNode;
        Node tempNode2 = null;
        try {
          outer:
            while (tempNode1 != null) {
                try {
                    this.visitPre(tempNode1);  // Always pre-visit
                    tempNode2 = tempNode1.getFirstChild();
                } catch (ToNextSiblingTraversalException exception) {
                    tempNode2 = null;           // Treat this node like a leaf.
                }
                if (tempNode2 == null) {        // If current node is a leaf node..
                    //if (!(tempNode1.equals(startNode))) {       // If not moving past stop node.. 
                    if (tempNode1 != startNode) {               // If not moving past stop node.. 
                        tempNode2 = tempNode1.getNextSibling(); // ..get leaf node's sibling 
                        if (tempNode2 != null) {                // If more siblings..
                            try {
                                this.visitPost(tempNode1);     // ..post visit this leaf.
                            } catch (ToNextSiblingTraversalException exception) {
                                // Do nothing because we're a leaf.
                            }
                        }
                    } else {                    // This leaf is the final node.
                        try {
                            this.visitPost(tempNode1); // ..post visit final node.
                        } catch (ToNextSiblingTraversalException exception) {
                            // Do nothing because we're finished anyway.
                        }
                        break outer;                                             
                    }
                } // Current node is a leaf node
                while (tempNode2 == null && tempNode1 != null) {// While leaf node is the last sibling..
                    tempNode2 = tempNode1.getParentNode();      // Get last leaf's parent
                    try {
                        this.visitPost(tempNode1);             // Post visit last leaf
                    } catch (ToNextSiblingTraversalException exception) {
                        // Do nothing because we're moving to our parent.
                    }
                    tempNode1 = tempNode2;
                    if (tempNode1 != null) {                    // If not at document root..
                        //if (!(tempNode1.equals(startNode))) {   // If not moving past stop node..
                        if (tempNode1 != startNode) {           // If not moving past stop node..
                            tempNode2 = tempNode1.getNextSibling(); // ..get parent's sibling.
                            if (tempNode2 != null) {            // If more siblings..
                                try {
                                    this.visitPost(tempNode1); // ..post visit this node.
                                } catch (ToNextSiblingTraversalException exception) {
                                    // Do nothing because we're moving to next sibling.
                                }
                            }
                        } else {                                // This node is the final node.
                            try {
                                this.visitPost(tempNode1);     // ..post visit final node.
                            } catch (ToNextSiblingTraversalException exception) {
                                // Do nothing because we're finished anyway.
                            }
                            break outer;        
                        }
                    }
                } 
                tempNode1 = tempNode2;
            } // outer: while tempNode1 != null
        } catch (EndTraversalException exception) {
            //  Suppress propogation of this condition
        }
    }

    /**
     *
     */
    private void visitPre(Node node) throws Exception {
        switch (node.getNodeType()) {
          case Node.ELEMENT_NODE:
            this.visitor.visitElementPre((Element)node);
            break;
          case Node.ATTRIBUTE_NODE:
            this.visitor.visitAttributePre((Attr)node);
            break;
          case Node.TEXT_NODE:
            this.visitor.visitTextPre((Text)node);
            break;
          case Node.CDATA_SECTION_NODE:
            this.visitor.visitCDATASectionPre((CDATASection)node);
            break;
          case Node.ENTITY_REFERENCE_NODE:
            this.visitor.visitEntityReferencePre((EntityReference)node);
            break;
          case Node.ENTITY_NODE:
            this.visitor.visitEntityPre((Entity)node);
            break;
          case Node.PROCESSING_INSTRUCTION_NODE:
            this.visitor.visitProcessingInstructionPre((ProcessingInstruction)node);
            break;
          case Node.COMMENT_NODE:
            this.visitor.visitCommentPre((Comment)node);
            break;
          case Node.DOCUMENT_NODE:
            this.visitor.visitDocumentPre((Document)node);
            break;
          case Node.DOCUMENT_TYPE_NODE:
            this.visitor.visitDocumentTypePre((DocumentType)node);
            throw new ToNextSiblingTraversalException(); // For incompatibility of XML4J DTD class
          case Node.DOCUMENT_FRAGMENT_NODE:
            this.visitor.visitDocumentFragmentPre((DocumentFragment)node);
            break;
          case Node.NOTATION_NODE:
            this.visitor.visitNotationPre((Notation)node);
            break;
          default:
            this.visitor.visitUnknownPre(node);
            break;
        }
    }

    /**
     *
     */
    private void visitPost(Node node) throws Exception {
        switch (node.getNodeType()) {
          case Node.ELEMENT_NODE:
            this.visitor.visitElementPost((Element)node);
            break;
          case Node.ATTRIBUTE_NODE:
            this.visitor.visitAttributePost((Attr)node);
            break;
          case Node.TEXT_NODE:
            this.visitor.visitTextPost((Text)node);
            break;
          case Node.CDATA_SECTION_NODE:
            this.visitor.visitCDATASectionPost((CDATASection)node);
            break;
          case Node.ENTITY_REFERENCE_NODE:
            this.visitor.visitEntityReferencePost((EntityReference)node);
            break;
          case Node.ENTITY_NODE:
            this.visitor.visitEntityPost((Entity)node);
            break;
          case Node.PROCESSING_INSTRUCTION_NODE:
            this.visitor.visitProcessingInstructionPost((ProcessingInstruction)node);
            break;
          case Node.COMMENT_NODE:
            this.visitor.visitCommentPost((Comment)node);
            break;
          case Node.DOCUMENT_NODE:
            this.visitor.visitDocumentPost((Document)node);
            break;
          case Node.DOCUMENT_TYPE_NODE:
            this.visitor.visitDocumentTypePost((DocumentType)node);
            break;
          case Node.DOCUMENT_FRAGMENT_NODE:
            this.visitor.visitDocumentFragmentPost((DocumentFragment)node);
            break;
          case Node.NOTATION_NODE:
            this.visitor.visitNotationPost((Notation)node);
            break;
          default:
            this.visitor.visitUnknownPost(node);
            break;
        }
    }

    // ================================================================
    /**
     * XML4J tree traversal exception which signals from a Visitor to the tree traversal algorithm
     * that further traversal of the current Node's children should be ended because the visit on 
     * this portion of thee document object hierarchy has been successfully completed.  For
     * example, the MakeDigestVisitor throws this exception when a digest already exists on a
     * given Node, and there is no need to traverse children of this Node.  Note that this exception
     * constitutes a request that a traversal algorithm is free to ignore.
     *
     * @author Tom Rowe &lt;trowe@us.ibm.com&gt;
     * @see com.ibm.xml.parser.ToNextSiblingTraversalException
     */
    public static class ToNextSiblingTraversalException extends Exception {
        /**
         * Constructor for exception with no detail message.
         */
        public ToNextSiblingTraversalException() {
            super();
        }
        /**
         * Constructor for exception with detail message.
         * @param msg       The detail message for the exception.
         */
        public ToNextSiblingTraversalException(String msg) {
            super(msg);
        }
    }

    // ================================================================
    /**
     * XML4J tree traversal exception which signals from a Visitor to the tree traversal algorithm
     * that all further traversal should be ended because the visit on the document object
     * hierarchy has been successfully completed.  For example, this exception could be used
     * by search Visitors that are looking for the first TXElement Node which contains certain 
     * state data.
     *
     * @author Tom Rowe &lt;trowe@us.ibm.com&gt;
     * @see com.ibm.xml.parser.EndTraversalException
     */
    public static class EndTraversalException extends Exception {

        /**
         * Constructor for exception with no detail message.
         */
        public EndTraversalException() {
            super();
        }

        /**
         * Constructor for exception with detail message.
         * @param msg       The detail message for the exception.
         */
        public EndTraversalException(String msg) {
            super(msg);
        }
    }
    // ================================================================
}

