/*
 * (C) Copyright IBM Corp. 1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.dsig;

import org.w3c.dom.*;
import org.xml.sax.helpers.ParserFactory;
import org.xml.sax.Parser;
import org.xml.sax.SAXException;
import com.ibm.xml.parsers.DOMParser;
import com.ibm.dom.util.Visitor;
import com.ibm.dom.util.ToXMLVisitor;
import com.ibm.dom.util.TreeTraversal;
import com.ibm.dom.util.Digest;

import java.security.*;
import java.io.*;
import java.util.*;
import java.net.*;
import java.security.Signature;
import java.security.cert.X509Certificate;

/**
 * @author YAGISHITA Kazuyo &lt;kazuyoy@jp.ibm.com&gt;
 */
public class XDSSign {
    
    public static void main(String[] argv) {
        
        String SaxParserClassName = "com.ibm.xml.parsers.SAXParser"; //default
        String DocumentClassName = "com.ibm.xml.dom.DocumentImpl"; //default

        if (argv.length < 2) {
            System.err.print("Usage: java XDSSign Resource SecurityContext ");
            System.err.print("(OutputFile Name) ");
            System.err.print("(SAXParserClass Name) ");
            System.err.println("(DocumentClass Name)");
            System.exit(1);
        }
        
        try {
            PrintWriter pw;
            if (argv.length < 3) {
                pw = new PrintWriter(System.out);
            } else {
                new File(argv[2]).delete();
                pw = new PrintWriter(new FileWriter(argv[2]));
                if (argv.length >= 4) {
                    SaxParserClassName = argv[3];
                    if(argv.length > 4) {
                        DocumentClassName = argv[4];
                    }
                }
            }

            XDSSign signer = new XDSSign(argv[1], SaxParserClassName);
            Document doc = signer.sign(argv[0], DocumentClassName);
            
            Visitor visitor = new ToXMLVisitor(pw);
            new TreeTraversal(visitor).traverse(doc);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    String id;
    String keyStorePath;
    String keyStorePassword;
    String keyPassword;
    String digestAlgorithm;
    String signatureAlgorithm;
    
    XDSSign(String securityContext, String SaxParserClassName) {
        
        try {
            Parser saxparser = ParserFactory.makeParser(SaxParserClassName);
            XDSSecurityContextHandler handler = new XDSSecurityContextHandler();
            saxparser.setDocumentHandler(handler);
            saxparser.parse(securityContext);
           
            this.id = handler.id;
            this.keyStorePath = handler.keyStorePath;
            this.keyStorePassword = handler.keyStorePassword;
            this.keyPassword = handler.keyPassword;
            this.digestAlgorithm = handler.digestAlgorithm;
            this.signatureAlgorithm = handler.signatureAlgorithm;
            
        } catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    public Document sign(String filename, String DocumentClassName) {
        Document doc = null;
        
        try {
            doc = (Document)Class.forName(DocumentClassName).newInstance();
            XDSElement root = new XDSElement(doc);
            
            XDSAlgorithm da = new XDSAlgorithm(doc, digestAlgorithm);
            da.digestAlgorithmID = digestAlgorithm;
            
            XDSPackageResource pr;
            byte[] value;
            if ( filename.endsWith(".xml") ) {
                DOMParser domparser = new DOMParser();
                try {    
                    domparser.parse(filename);
                } catch (SAXException se) {    
                    se.printStackTrace();
                } catch (IOException ioe) {
                    ioe.printStackTrace();
                }
                Element res = domparser.getDocument().getDocumentElement();
                pr = new XDSPackageResource(doc, res, digestAlgorithm, "data");
                MessageDigest md = MessageDigest.getInstance("MD5");
                value = Digest.getDigest(md, res);
            } else {
                BufferedReader br = new BufferedReader(new FileReader(filename));
                String nextLine = null;
                String line = "";
                while ((nextLine = br.readLine()) != null) {
                    line = line + nextLine;
                }
                byte[] b = line.getBytes();
                pr = new XDSPackageResource(doc, b, digestAlgorithm, "data");
                pr.value.encoding = "base64";
                MessageDigest md = MessageDigest.getInstance(digestAlgorithm);
                value = md.digest(b);
            }
            
            root.setPackage(da, pr);
            
            XDSSignatureResource resource = new XDSSignatureResource(doc, "#data", da, value);
            resource.digest.value.encoding = "base64";
            
            KeyStore keyStore = KeyStore.getInstance("jks");
            keyStore.load(new FileInputStream(keyStorePath),keyStorePassword.toCharArray());
            X509Certificate cert = (X509Certificate)keyStore.getCertificate(id);
      
            XDSInfo originatorInfo = new XDSInfo(doc);
            XDSIssuerAndSerialNumber iasn 
                = new XDSIssuerAndSerialNumber(doc, 
                                               cert.getIssuerDN().toString(), 
                                               cert.getSerialNumber().toString());
            originatorInfo.addInformation(iasn);
            
            XDSAlgorithm sa = new XDSAlgorithm(doc, signatureAlgorithm);
            
            XDSManifest manifest = new XDSManifest(doc);
            manifest.resource = resource;
            manifest.originator = originatorInfo;
            manifest.signatureAlgorithm = sa;
            
            XDSValue signedValue = new XDSValue(doc);
            signedValue.encoding = "base64";
            
            root.addSignature(manifest, signedValue);
            doc.appendChild(root.getDSigDocument());
            
            // Sign it
            NodeList nl = doc.getElementsByTagName("dsig:Signature");
            
            for ( int i=0; i<nl.getLength(); i++ ) {
                Element manifest_element 
                    = (Element)(nl.item(i)).getFirstChild();
                Element value_element 
                    = (Element)(nl.item(i)).getLastChild();
                Key mykey = keyStore.getKey(id,keyPassword.toCharArray());
                if (mykey == null) {
                    System.err.println("Specified key not found in the keyStore.");
                    System.exit(-1);
                }
                Signature sig = Signature.getInstance(signatureAlgorithm);
                sig.initSign((PrivateKey)mykey);
                MessageDigest md = MessageDigest.getInstance("MD5");
                byte[] digest = Digest.getDigest(md, manifest_element);
                sig.update(digest);
                if ( signedValue.encoding.equals("base64") ) {
                    value_element.setAttribute("encoding", "base64");
                    value_element.appendChild(doc.createTextNode
                                              (signedValue.getBase64Digest(sig.sign())));
                } else {
                    value_element.setAttribute("encoding", "none");
                    value_element.appendChild(doc.createTextNode(new String(sig.sign())));
                }
            }
            
        } catch (Exception e) {
            e.printStackTrace();
        }
        
        return doc;
        
    }

}
