/*
 * (C) Copyright IBM Corp. 1999  All rights reserved.
 *
 * US Government Users Restricted Rights Use, duplication or
 * disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * IBM will not be liable for any damages suffered by you as a result
 * of using the Program. In no event will IBM be liable for any
 * special, indirect or consequential damages or lost profits even if
 * IBM has been advised of the possibility of their occurrence. IBM
 * will not be liable for any third party claims against you.
 */

package com.ibm.xml.sax;

import java.io.CharArrayWriter;
import java.io.IOException;
import java.io.Writer;
import java.security.MessageDigest;
import java.util.Hashtable;
import java.util.Stack;
import java.util.Vector;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;

/**
 *
 * @version Revision: %M% %I% %W% %Q%
 * @author TAMURA Kent &lt;kent@trl.ibm.co.jp&gt;
 */
public class DigestFilter implements NamespaceDocumentHandler {
    protected CharArrayWriter charBuffer = null;
    protected MessageDigest messageDigest;
    protected Stack elementStack = null;
    protected byte[] result;
    protected NamespaceDocumentHandler next = null;
    protected boolean enable = false;

    /**
     * Creates a <CODE>DigestFilter</CODE> instance.
     */
    public DigestFilter() {
    }

    /**
     * Creates a <CODE>DigestFilter</CODE> instance.
     */
    public DigestFilter(NamespaceDocumentHandler next) {
        this.next = next;
    }

    /**
     * Creates a <CODE>DigestFilter</CODE> instance in `enable' mode.
     */
    public DigestFilter(MessageDigest md) {
        this.enable(md);
    }

    /**
     * Starts digest calculation.
     */
    public void enable(MessageDigest md) {
        if (this.enable) {
            if (!this.messageDigest.getAlgorithm().equals(md.getAlgorithm()))
                throw new RuntimeException("DigestFilter#enable(): Don't use a MessageDigest different from previous one.");
        } else {
            this.enable = true;
            this.messageDigest = md;
        }
        if (this.charBuffer == null)
            this.charBuffer = new CharArrayWriter();
        if (this.elementStack == null)
            this.elementStack = new Stack();
    }

    /**
     * Returns a digest value.
     */
    public byte[] getDigest() {
        return result;
    }

    /**
     *
     */
    public void setNextDocumentHandler(NamespaceDocumentHandler handler) {
        this.next = handler;
    }

    /**
     * Applications must not call this method.
     */
    public void startElement(String namespace, String prefix, String localPart,
                             String qname, NamespaceAttributeList atts)
        throws SAXException {
        this.flushCharBuffer();

        if (this.next != null)
            this.next.startElement(namespace, prefix, localPart, qname, atts);
        if (!this.enable)
            return;

        int atlen = atts.getLength();
        int lenWithoutXmlns = 0;
        Hashtable hash = new Hashtable(atlen);
        String[] as = new String[atlen];        // Attribute digests orderred by expanded name
        StringBuffer sb = new StringBuffer();
        for (int i = 0;  i < atlen;  i ++) {
            String p = atts.getPrefix(i);
            if (!(atts.getName(i).equals("xmlns") || (p != null && p.equals("xmlns")))) {
                String ns = atts.getNamespace(i);
                String lp = atts.getLocalPart(i);
                String expandedName;
                if (ns == null) {
                    expandedName = lp.length() == 0 ? ":" : lp;
                } else {
                    sb.setLength(0);
                    sb.append(ns);
                    sb.append(":");
                    sb.append(lp);
                    expandedName = sb.toString();
                }
                hash.put(expandedName, atts.getName(i));
                as[lenWithoutXmlns++] = expandedName;
            }
        }
        Object[] digestArray = new Object[lenWithoutXmlns]; // byte[][]
        heapSort(as, lenWithoutXmlns);
        for (int i = 0;  i < lenWithoutXmlns;  i ++) {
            String ename = as[i];
            String value = atts.getValue((String)hash.get(as[i]));
            //System.err.println("\""+com.ibm.xml.parser.TXText.makePrintable(value)+"\"");
            this.messageDigest.reset();
            digestInt(this.messageDigest, 2);   // org.w3c.dom.Node.ATTRIBUTE_NODE
            digestString(this.messageDigest, ename);
            this.messageDigest.update((byte)0);
            this.messageDigest.update((byte)0);
            digestString(this.messageDigest, value);
            digestArray[i] = this.messageDigest.digest();
        }
        ElementDigest edigest = new ElementDigest(digestArray);
        this.elementStack.push(edigest);
    }

    /**
     * Applications must not call this method.
     */
    public void endElement(String namespace, String prefix, String localPart, String qname)
        throws SAXException {
        this.flushCharBuffer();

        if (this.enable) {
            ElementDigest edigest = (ElementDigest)this.elementStack.pop();
            this.messageDigest.reset();
            digestInt(this.messageDigest, 1);   // org.w3c.dom.Node.ELEMENT_NODE
            if (namespace != null) {
                digestString(this.messageDigest, namespace);
                digestString(this.messageDigest, ":");
            }
            digestString(this.messageDigest, localPart);
            this.messageDigest.update((byte)0);
            this.messageDigest.update((byte)0);
            digestInt(this.messageDigest, edigest.attrDigests.length);
            for (int i = 0;  i < edigest.attrDigests.length;  i ++)
                this.messageDigest.update((byte[])edigest.attrDigests[i]);
            digestInt(this.messageDigest, edigest.getChildrenSize());
            for (int i = 0;  i < edigest.getChildrenSize();  i ++)
                this.messageDigest.update((byte[])edigest.childrenDigests.elementAt(i));
            this.recordDigest(this.messageDigest.digest());

            if (this.elementStack.size() == 0) {
                this.enable = false;
                this.messageDigest = null;
            }
        }
        if (this.next != null)
            this.next.endElement(namespace, prefix, localPart, qname);
    }

    /**
     * Applications must not call this method.
     */
    public void characters(char[] ch, int start, int length)
        throws SAXException {
        if (this.next != null)
            this.next.characters(ch, start, length);
        if (this.enable)
            this.appendCharBuffer(ch, start, length);
    }

    /**
     * Applications must not call this method.
     */
    public void ignorableWhitespace(char[] ch, int start, int length)
        throws SAXException {
        if (this.next != null)
            this.next.ignorableWhitespace(ch, start, length);
        if (this.enable)
            this.appendCharBuffer(ch, start, length);
    }

    /**
     * Applications must not call this method.
     */
    public void processingInstruction(String target, String data)
        throws SAXException {
        this.flushCharBuffer();
        if (this.enable) {
            this.messageDigest.reset();
            digestInt(this.messageDigest, 7);   // org.w3c.dom.Node.PROCESSING_INSTRUCTION_NODE
            digestString(this.messageDigest, target);
            this.messageDigest.update((byte)0);
            this.messageDigest.update((byte)0);
            int start = 0;                      // Skip S
            for (;  start < data.length();  start ++) {
                int ch = data.charAt(start);
                if (ch != ' ' && ch != '\t' && ch != '\r' && ch != '\n')
                    break;
            }
            if (start == 0) {
                digestString(this.messageDigest, data);
            } else if (start < data.length()) {
                digestString(this.messageDigest, data.substring(start));
            }
            this.recordDigest(this.messageDigest.digest());
        }
        if (this.next != null)
            this.next.processingInstruction(target, data);
    }

    private void appendCharBuffer(char[] ch, int start, int length) {
        this.charBuffer.write(ch, start, length);
    }
    private void flushCharBuffer() {
        if (this.enable && this.charBuffer.size() > 0) {
            this.messageDigest.reset();
            digestInt(this.messageDigest, 3);   // org.w3c.dom.Node.TEXT_NODE
            char[] carr = this.charBuffer.toCharArray();
            for (int i = 0;  i < carr.length;  i ++) {
                int ch = carr[i];
                this.messageDigest.update((byte)(ch >> 8));
                this.messageDigest.update((byte)(ch & 0xff));
            }
            /*
            try {
            System.err.print(carr.length+",");
            //System.err.println("\""+com.ibm.xml.parser.TXText.makePrintable(new String(carr))+"\"");
            printByteArray(new String(carr).getBytes("UnicodeBigUnmarked"));
            } catch (Exception ex) {
            }
            */
            this.recordDigest(this.messageDigest.digest());
            this.charBuffer.reset();
        }
    }

    private void recordDigest(byte[] digest) {
                                        // append this digest value to current Element.
                                        // When there is no current Element
                                        // -- it means this method was called in endElement()
                                        // for the root element --,
                                        // the digest value is stored as a result of this process.
        this.result = digest;
        if (!this.elementStack.empty()) {
            ElementDigest edigest = (ElementDigest)this.elementStack.peek();
            edigest.addDigest(digest);
        }
        /*
        try {
            printByteArray(digest);
        } catch (java.io.IOException io) {
        }
        */
    }

    /**
     * Writes 32bit integer in network byte order.
     */
    private static void digestInt(MessageDigest md, int i) {
        md.update((byte)((i >> 24) & 0xff));
        md.update((byte)((i >> 16) & 0xff));
        md.update((byte)((i >> 8) & 0xff));
        md.update((byte)(i & 0xff));
    }

    /**
     * Writes String in big endian UTF-16.
     */
    private static void digestString(MessageDigest md, String str) {
        for (int i = 0;  i < str.length();  i ++) {
            int ch = str.charAt(i);
            md.update((byte)(ch >> 8));
            md.update((byte)(ch & 0xff));
        }
    }

    private static class ElementDigest {
        Object[] attrDigests;
        Vector childrenDigests = null;
        ElementDigest(Object[] ad) {
            this.attrDigests = ad;
        }
        void addDigest(byte[] digest) {
            if (this.childrenDigests == null)
                this.childrenDigests = new Vector();
            this.childrenDigests.addElement(digest);
        }
        int getChildrenSize() {
            return this.childrenDigests == null ? 0 : this.childrenDigests.size();
        }
    }

    /**
     * Applications must not call this method.
     */
    public void setDocumentLocator(Locator locator) {
        if (this.next != null)  this.next.setDocumentLocator(locator);
    }

    /**
     * Applications must not call this method.
     */
    public void startDocument() throws SAXException {
        if (this.next != null)  this.next.startDocument();
    }

    /**
     * Applications must not call this method.
     */
    public void endDocument() throws SAXException {
        if (this.next != null)  this.next.endDocument();
    }


    /**
     * Sort String array.
     */
    public static void heapSort(String[] pd, int length) {
        int i;
        for (i = length/2;  i >= 0;  i--) {     // Make heap
            fall(pd, length, i);
        }
        for (i = length-1;  i > 0;  i--) {
            String t = pd[0];
            pd[0] = pd[i];
            pd[i] = t;
            fall(pd, i, 0);
        }
    }
    static private void fall(String[] pd, int n, int i) {
        int j = 2*i+1;
        if (j < n) {                            // left exists
            if (j+1 < n) {                      // right exists too
                                                // j: bigger
                if (0 > pd[j].compareTo(pd[j+1]))
                    j = 2*i+2;
            } else {                            // only left
            }
            if (0 > pd[i].compareTo(pd[j])) {
                                                // the child is bigger
                String t = pd[i];
                pd[i] = pd[j];
                pd[j] = t;
                fall(pd, n, j);
            }
        }
    }

    public static void printByteArray(byte[] ab) throws IOException {
        if (null == ab) {
            System.err.print("NULL");
        } else {
            for (int i = 0;  i < ab.length;  i ++) {
                System.err.print(tab.charAt((ab[i]>>4)&0x0f));
                System.err.print(tab.charAt(ab[i]&0x0f));
            }
        }
        System.err.print("\n");
    }
    public static void printByteArray(Writer wr, byte[] ab) throws IOException {
        if (null == ab) {
            wr.write("NULL");
        } else {
            for (int i = 0;  i < ab.length;  i ++)
                printByte(wr, ab[i]);
        }
        wr.write("\n");
    }
    static final String tab = "0123456789ABCDEF";
    private static void printByte(Writer wr, byte b) throws IOException {
        wr.write(tab.charAt((b>>4)&0x0f));
        wr.write(tab.charAt(b&0x0f));
    }
}
